/* maps2 -- convert a multiple alignment ("refine" output) to PostScript */

const char rcsid[] =
"$Id: maps2.c,v 1.27 2004/05/20 01:01:43 webb Exp $";

#include "util.h"
#include "seq.h"
#include "exons.h"
#include "args.h"
#include "piplib.h"
#include "mlav.h"
#include "cmap.h"

#define MAX_SEQ 100000
#define MAX_ROW 100
#define MAX_PAGE 100000
#define TOP_ROW 700
#define BOTTOM_ROW 50
#define PTS_PER_ROW 10
#define BETWEEN_ROWS 30
#define FEATURES_ROW 20
#define ANN_SEP 5
#define EACH_ANN 3

typedef struct feat {
	int beg, end, type, page;
	char *title;
} Feature;

typedef struct gene {
	int beg, end, first_exon, nexon, page;
	char *name;
} Gene;

typedef struct type {		/* kind of entry in the annotation file */
	char *name, *color;
} Type;

typedef struct ann {		/* kind of entry in the annotation file */
	int beg, end, ordinal, row;
	char *type, *color;
} Annotation;

/* overlap - does [b1,e1] overlap [b2,e2]? */
int overlap(int b1, int e1, int b2, int e2)
{
	return (e1 >= b2 && b1 <= e2);
}

int str_is_safechars(const char* s)
{
    if (!s) return 0;
    for (; *s; ++s)
	if (! (isalnum(*s) || (*s == '_') ))
	    return 0;
    return 1;
}

void check_safechars(const char* s)
{
    if (!str_is_safechars(s))
	fatalf("the string {%s} contains invalid characters", s);
}

/* layout - stack overlapping annotations into rows */
int layout(int b, int e, Annotation *a, int nann)
{
	int i, j, k, r, nann_row = 0, taken[10000], max_taken;

	for (i = 0; i < nann; ++i) {
		if (overlap(b, e, a[i].beg, a[i].end)) {
			/* find the row for annotation i */
			max_taken = taken[0] = 0;
			for (j = 0; j < i; ++j)
			    if ((r = a[j].row) > 0 &&
			      overlap(a[j].beg, a[j].end, a[i].beg, a[i].end)) {
					if (r > max_taken) {
					    for (k = max_taken+1; k < r; ++k)
						taken[k] = 0;
					    taken[r] = 1;
					    max_taken = r;
					}
			}
			for (r = 1; r <= max_taken; ++r)
				if (taken[r] == 0)
					break;
			a[i].row = r;
			nann_row = MAX(nann_row, r);
		} else
			a[i].row = 0;
	}
	return nann_row;
}

/* return row 0 and the array of corresponding positions in the reference seq */
void row0(Mseq *mseq, char *row, int *col2pos)
{
	int j;

	for (j = 0; j < 50; ++j) {
		if (mseq->scan == NULL) {
			row[j] = ' ';
			col2pos[j] = 0;
			continue;
		} else if (mseq->scan->size > 0) {
			row[j] = mseq->s[mseq->cur-1];
			col2pos[j] = mseq->cur++;
			mseq->scan->size--;
		} else {
			row[j] =
			   (mseq->cur == mseq->beg ||
			    mseq->cur == mseq->end + 1)
			  	? ' ' : '-';
			col2pos[j] = 0;
			mseq->scan->size++;
		}
		if (mseq->scan->size == 0)
			mseq->scan = mseq->scan->next;
	}
}

/* print out a row other than row 0 */
void print_row(char *row0, Mseq *mseq)
{
	int j;
	char c;

	putchar('(');
	for (j = 0; j < 50; ++j) {
		if (mseq->scan == NULL) {
			putchar(' ');
			continue;
		} else if (mseq->scan->size > 0) {
			c = mseq->s[mseq->cur-1];
			putchar(c != row0[j] ? c : '.');
			mseq->scan->size--;
			mseq->cur++;
		} else {
			putchar((mseq->cur == mseq->beg || mseq->cur == mseq->end+1)
			   ? ' ' : '-');
			mseq->scan->size++;
		}
		if (mseq->scan->size == 0) {
			mseq->scan = mseq->scan->next;
			if (mseq->scan && mseq->scan->size < 0 &&
			    mseq->scan->next == NULL)
				mseq->scan = NULL;
		}
	}
	printf(")\n");
}

/* pos2col - map sequence position into column index */
void pos2col(int b, int e, int begpos, int endpos, int *col2pos, int *beg_col,
  int *end_col)
{
	int j, cb, ce;

	if (b < begpos)
		cb = 1;
	else {
		cb = 0;
		for (j = 0; j < 50; ++j)
			if (col2pos[j] == b)
				cb = j+1;
		if (cb == 0)
			fatal("pos2col: cannot find start.");
	}
	if (e > endpos)
		ce = 50;
	else {
		ce = 0;
		for (j = 0; j < 50; ++j)
			if (col2pos[j] == e)
				ce = j+1;
		if (ce == 0)
			fatal("pos2col: cannot find end.");
	}
	*beg_col = cb;
	*end_col = ce;

}

void show_annot(Annotation *a, int nann, int *col2pos, int begpos, int endpos,
  int features)
{
	int i, cb, ce;

	for (i = 0; i < nann; ++i) {
		if (a[i].row == 0)
			continue;
		pos2col(a[i].beg, a[i].end, begpos, endpos, col2pos, &cb, &ce);
		printf("gsave %d %d %d %d (annot.%d) %s annot grestore\n",
		    cb, ce, features, a[i].row, a[i].ordinal, a[i].color);
	}
}

void show_underlay(Feature *underlay, int nund, int nrow, int *col2pos,
  int begpos, int endpos)
{
	int i, b, e, t, cb, ce;

	for (i = 0; i < nund; ++i) {
		b = underlay[i].beg;
		e = underlay[i].end;
		t = underlay[i].type;
		if (overlap(b, e, begpos, endpos)) {
		      pos2col(b, e, begpos, endpos, col2pos, &cb, &ce);
		      printf("gsave\n%s %d %d %d %s_half colorbox\ngrestore\n",
			underlay[i].title, cb, ce, nrow,
			(t == 0 ? "top" : (t == 1 ? "bottom" : "both")));
		}
	}
}

void show_conserved(Feature *conserved, int ncon, int nrow, int *col2pos,
  int begpos, int endpos)
{
	int i, b, e, cb, ce;

	for (i = 0; i < ncon; ++i) {
		b = conserved[i].beg;
		e = conserved[i].end;
		if (overlap(b, e, begpos, endpos)) {
		      pos2col(b, e, begpos, endpos, col2pos, &cb, &ce);
		      printf("%d %d %d box\n", cb, ce, nrow);
		}
	}
}

void show_features(Feature *feature, int nfeat, int *col2pos, int begpos,
 int endpos)
{
	int i, b, e, cb, ce, t;

	for (i = 0; i < nfeat; ++i) {
		b = feature[i].beg;
		e = feature[i].end;
		if (overlap(b, e, begpos, endpos)) {
			pos2col(b, e, begpos, endpos, col2pos, &cb, &ce);
			t = feature[i].type;
			printf("(%s) %d %d %sarrow\n", feature[i].title, cb,ce,
			  t == 1 ? "right" : (t == 0 ? "non" : "left"));
		}
	}
}

void show_align(Mseq *mseq, int nseq, Annotation *a, int nann, Gene *gene,
  int ngene, Feature *exon, int nexon, Feature *repeat, int nrpt,
  Feature *underlay, int nund, Feature *conserved, int ncon, FILE *bp)
{
	int y, nrow, start_pos[MAX_ROW], i, j, lim, col2pos[50], drop, page,
	  first_pos[MAX_PAGE], begpos, endpos, nann_rows, features, top_stuff;
	char row[50], *seqname[MAX_ROW];
	Block_ptr ss;

	printf("[/Title (alignment) /Dest [{ThisPage} /FitB] /OUT pdfmark\n");
	page = y = 0;
	while  (mseq[0].scan) {
		start_pos[0] = mseq[0].cur;
		for (nrow = i = 1; i < nseq; ++i)
			if (mseq[i].scan &&
			   (mseq[i].scan!=mseq[i].block || mseq[i].scan->size>-50))
				start_pos[nrow++] = mseq[i].cur;
		row0(mseq, row, col2pos);
		begpos = endpos = 0;
		for (i = 0; i < 50; ++i)
			if ((j = col2pos[i]) > 0) {	/* what if all 0? */
				if (begpos == 0)
					begpos = j;
				endpos = j;
			}
		nann_rows = layout(begpos, endpos, a, nann);
		
		features = 0;
		for (i = 0; i < nexon; ++i)
			if (exon[i].beg <= endpos && exon[i].end >= begpos)
				features = 1;
		for (i = 0; i < nrpt; ++i)
			if (repeat[i].beg <= endpos && repeat[i].end >= begpos)
				features = 1;
		drop = PTS_PER_ROW*nrow + BETWEEN_ROWS;
		top_stuff = FEATURES_ROW*features;
		if (nann_rows)
			top_stuff += (ANN_SEP + nann_rows*EACH_ANN);
		if (y - drop - top_stuff < BOTTOM_ROW) {
			if (page > 0)
				printf("pgsave restore\nshowpage\n");
			if (page >= MAX_PAGE)
				fatalf("error: more than %d pages of output",
					MAX_PAGE);
			first_pos[page++] = start_pos[0];
			printf("%%%%Page: %d %d\n", page, page);
			printf("%%%%BeginPageSetup\n/pgsave save def\n");
			printf("%%%%EndPageSetup\n0 %d translate\n", TOP_ROW);
			printf("[/Dest (page%d_) cvn ", page);
			printf("/View [/FitB] /DEST pdfmark\n");
			y = TOP_ROW;
		}

		if (top_stuff) {
			printf("\n0 -%d translate\n", top_stuff);
			y -= top_stuff;
		}
		
		if (nann_rows)
			show_annot(a, nann, col2pos, begpos, endpos, features);
		if (features) {
			printf("TimesRoman10 setfont\n");
			show_features(repeat, nrpt, col2pos, begpos, endpos);
			show_features(exon, nexon, col2pos, begpos, endpos);
		}
		for (i = 0; i < nexon; ++i)
			if (begpos <= exon[i].beg && exon[i].beg <= endpos)
				exon[i].page = page;
		for (i = 0; i < ngene; ++i)
			if (begpos <= gene[i].beg && gene[i].beg <= endpos)
				gene[i].page = page;
		printf("[\n");
		for (i = 9; i < 50; i += 10)
			if (col2pos[i] == 0)
				printf("()");
			else
				printf("(%d)", col2pos[i]);
		printf("\n] showpos\n");

		show_underlay(underlay, nund, nrow, col2pos, begpos, endpos);
		show_conserved(conserved, ncon, nrow, col2pos, begpos, endpos);
	
		printf("[\n");
		for (i = 0; i < nrow; ++i)
			printf("(%d)\n", start_pos[i]);
		printf("] column1\n");

		printf("[\n(");
		for (j = 0; j < 50; ++j)
			putchar(row[j]);
		printf(")\n");
	
		seqname[0] = mseq[0].name;
		for (j = i = 1; i < nseq; ++i)
			if ((ss = mseq[i].scan) == NULL)
				continue;
			else if (ss == mseq[i].block && ss->size <= -50) {
				if ((ss->size += 50) == 0)
					mseq[i].scan = mseq[i].scan->next;
			} else  {
				seqname[j++] = mseq[i].name;
				print_row(row, mseq + i);
			}

		printf("] column2\n[\n");
		for (i = 0; i < nrow; ++i)
			printf("(%s)\n", seqname[i]);
		printf("] column3\n\n");

		printf("\n0 -%d translate\n", drop);
		y -= drop;
	}

	printf("pgsave restore\nshowpage\n");

	
	fprintf(bp, "[/Page 1 /View [/FitB] /Title (genes) ");
	fprintf(bp, "/Count %d /OUT pdfmark\n", ngene);
	for (i = 0; i < ngene; ++i) {
		fprintf(bp, "[/Dest (page%d_) cvn /View [/FitB] /Title (%s) ",
		  gene[i].page, gene[i].name);
		fprintf(bp, "/Count %d /OUT pdfmark\n", -gene[i].nexon);
		lim = gene[i].first_exon + gene[i].nexon;
		for (j = gene[i].first_exon; j < lim; ++j) {
			fprintf(bp, "[/Dest (page%d_) cvn /View [/FitB] ",
			  exon[j].page);
			fprintf(bp, "/Title (%d %d) /OUT pdfmark\n",
			  exon[j].beg, exon[j].end);
		}

	}

	fprintf(bp, "[/Page 1 /View [/FitB] /Title (Position Index) ");
	fprintf(bp, "/Count %d /OUT pdfmark\n", -page);
	for (i = 0; i < page; ++i) {
		fprintf(bp,"[/Dest (page%d_) cvn /View [/FitB] /Title (%d) ",
		  i+1, first_pos[i]);
		fprintf(bp, "/OUT pdfmark\n");
	}

	fprintf(bp, "%%%%Trailer\n%%%%Pages: %d\n", page); /* XXX */
}

int read_annotations(char *filename, Annotation *a)
{
	int nann = 0, ntype = 0, i, b, e, done, order = 0;
	Type type[MAX_SEQ];
	FILE *fp = ckopen(filename, "r");
	char buf[500], string[500];

	if (fgets(buf, 500, fp) == NULL)
		return 0;
	for ( ; ; ) {
		if (strncmp(buf, "%define type", 11) == 0) {
			if (fgets(buf, 500, fp) == NULL)
				break;
			if (sscanf(buf, "%%name %499s", string) == 1) {
				check_safechars(string);
				type[ntype].name = copy_string(string);
				fgets(buf, 500, fp);
				if (sscanf(buf, "%%color %499s", string) == 1) {
					check_safechars(string);
					type[ntype].color = copy_string(string);
				} else {
					fatalf("color: %s", buf);
				}
			} else if (sscanf(buf, "%%color %499s", string) == 1) {
				check_safechars(string);
				type[ntype].color = copy_string(string);
				fgets(buf, 500, fp);
				if (sscanf(buf, "%%name %499s", string) == 1) {
					check_safechars(string);
					type[ntype].name = copy_string(string);
				} else {
					fatalf("name: %s", buf);
				}
			} else {
				fatalf("name or color: %s", buf);
			}
			++ntype;
		} else if (strncmp(buf, "%define annotation", 17) == 0) {
			for ( ; ; ) {
				if (fgets(buf, 500, fp) == NULL) {
					done = 1;
					break;
				}
				if (strncmp(buf, "%define ", 8) == 0) {
					done = 0;
					break;
				}
				if (sscanf(buf, "%%type %499s", string) == 1) {
				    check_safechars(string);
				    a[nann].type = copy_string(string);
				    for (i = 0; i < ntype; ++i)
					if (same_string(type[i].name, string)) {
						a[nann].color =
						     copy_string(type[i].color);
						break;
					}
				    if (i == ntype)
					fatalf("type %s undefined", string);
				} else if (sscanf(buf,"%%range %d %d", &b, &e)
				   == 2) {
					a[nann].beg = b;
					a[nann].end = e;
					a[nann].ordinal = order++;
				}
			}
			++nann;
			if (done)
				break;
		} else if (fgets(buf, 500, fp) == NULL)
			break;
	}
	
	for (i = 0; i < ntype; ++i) {
		free(type[i].name);
		free(type[i].color);
	}
	fclose(fp);
	return nann;
}

int read_exons(char *filename, Feature *exon, Gene *gene, int *ngene_ptr)
{
	int ngene, nexon, N, e, k, n, d, first_exon, incr, *beg, *end, *dir;
	char gname[500], ename[500], **gene_name, **exon_name;

	N = get_exons(filename, &beg, &end, &dir, &gene_name, &exon_name, 0);
	ngene = nexon = n = 0;
	while (n < N) {
		if (gene_name[n] == NULL)
			fatal("Cannot find gene name.");
		if (ngene >= MAX_SEQ)
			fatal("read_exons: too many genes");
		gene[ngene].name = copy_string(gene_name[n]);
		quote_for_ps(gene[ngene].name);
		gene[ngene].beg = beg[n];
		gene[ngene].end = end[n];
		strcpy(gname, gene_name[n]);
		d = dir[n];
		if (++n == N)
			fatalf("gene %s has no CDS information.",
				gene_name[n-1]);
		for (first_exon = ++n; n < N && gene_name[n] == NULL; ++n)
			;
		gene[ngene].first_exon = nexon;
		gene[ngene++].nexon = n - first_exon;
		if (d >= 0) {	/* on forward strand or no direction */
			e = 1;
			incr = +1;
		} else {
			e = n - first_exon;
			incr = -1;
		}
		
		for (k = first_exon; k < n; ++k, e += incr) {
			if (nexon >= MAX_SEQ)
				fatal("read_exons: too many exons");
			if (d == 0)
				strcpy(ename, gname);
			else if (exon_name[k] == NULL)
				sprintf(ename, "%s:%d", gname, e);
			else
				sprintf(ename, "%s:%s", gname, exon_name[k]);
			exon[nexon].beg = beg[k];
			exon[nexon].end = end[k];
			exon[nexon].type = d;
			exon[nexon].title = copy_string(ename);
			quote_for_ps(exon[nexon].title);
			++nexon;
		}
	}
	free_exons(beg, end, dir, gene_name, exon_name, N);
	*ngene_ptr = ngene;
	return nexon;
}

int read_conserved(char *filename, Feature *conserved)
{
	int ncon = 0, b, e;
	FILE *fp;
	char buf[500];

	fp = ckopen(filename, "r");
	while (fgets(buf, 500, fp))
		if (sscanf(buf, "%d %d", &b, &e) == 2) {
			if (ncon >= MAX_SEQ)
				fatal("read_conserved: too many conservons");
			conserved[ncon].beg = b;
			conserved[ncon].end = e;
			++ncon;
		}
	fclose(fp);
	return ncon;
}

int read_repeats(char *filename, Feature *repeat)
{
	int nrpt = 0, b, e, t;
	FILE *fp;
	char buf[500], title[500];

	fp = ckopen(filename, "r");
	while (fgets(buf, 500, fp)) {
		t = 0;
		if (sscanf(buf, "%d %d Left %499s", &b, &e, title) == 3)
			t = 1;
		else if (sscanf(buf, "%d %d Right %499s", &b, &e, title) == 3)
			t = 2;
		if (t > 0) {
			check_safechars(title);
			if (nrpt >= MAX_SEQ)
				fatal("read_repeats: too many repeats");
			repeat[nrpt].beg = b;
			repeat[nrpt].end = e;
			repeat[nrpt].type = t;
			repeat[nrpt].title = copy_string(title);
			++nrpt;
		}
	}
	fclose(fp);
	return nrpt;
}

int read_underlay(char *filename, Feature *underlay)
{
	int nund = 0, b, e;
	FILE *fp;
	char buf[500], color[500], name[500], *s, *last;

	fp = ckopen(filename, "r");
	while (fgets(buf, 500, fp)) {
		for (s = buf; *s && isspace(*s); ++s)
			;
		if (isdigit(*s) &&
		    sscanf(s, "%d %d %499s", &b, &e, color) == 3) {
			if (nund >= MAX_SEQ) { fatalf("too many underlays\n"); }
			underlay[nund].beg = b;
			underlay[nund].end = e;
			last = color + strlen(color) - 1;
			if (*last == '+') {
				underlay[nund].type = 0;
				*last = '\0';
			} else if (*last == '-') {
				underlay[nund].type = 1;
				*last = '\0';
			} else
				underlay[nund].type = 2;
			underlay[nund].title = cmap_get(color);
			++nund;
		} else if (isupper(*s) && 
		    sscanf(s, "%499s %499s", color, name) == 2) {
			check_basic_color(color);
			check_safechars(name);
			cmap_add(name, color);
		}
	}
	fclose(fp);
	return nund;
}

void cat(char *path)
{
	int ch;
	FILE *fp = fopen(path, "r");
	if (!fp)
	    fatalf("fopen(%s)", path);
	while ((ch = getc(fp)) != EOF)
	    if (putchar(ch) == EOF)
		fatalf("putc(%d)", ch);
	fclose(fp);
}


Feature exon[MAX_SEQ], repeat[MAX_SEQ], underlay[MAX_SEQ],
	  conserved[MAX_SEQ];
Gene gene[MAX_SEQ];
Annotation annotation[MAX_SEQ];

int main(int argc, char *argv[])
{
	int i, j, nseq, nann = 0, ngene = 0, nexon = 0, nrpt = 0, nund = 0,
	  ncon = 0;
	Mseq *mseq;
	char *filename, *procset, buf[500], *s, *t, oldroot[500], root[500],
	  substitute[500];
	FILE *bp;

	argv0 = argv[0];

	if (argc < 2)
	  fatal("args = refine-output-file seqname1 .. -- [ABCERUP=..]");

	cmap_init();

	nseq = read_mseqs(argv[1], &mseq);
	substitute[0] = oldroot[0] = '\0';
	for (i = 2, j = 0; j < nseq; ++j) {
		strcpy(root, mseq[j].name);
		if (root[strlen(root)-1] == '-')
			root[strlen(root)-1] = '\0';
		if ((t = strchr(root, ':')) != NULL)
			*t = '\0';
		if (strcmp(root, oldroot)) {
			strcpy(oldroot, root);
			bp = ckopen(argv[i], "r");
			if (fgets(substitute, 500, bp) == NULL)
				fatalf("%s is empty");
			quote_for_ps(substitute);
			for (s = substitute; *s; ++s)
				if (isspace(*s))
					break;
				*s = '\0';
			fclose(bp);
			++i;
		}
		free(mseq[j].name);
		if (t != NULL) {
			sprintf(buf, "%s:%s", substitute, t+1); 
			mseq[j].name = copy_string(buf);
		} else
			mseq[j].name = copy_string(substitute);
	}

	if (i >= argc)
		fatal("command-line arguments much contain '--'");

	if (strcmp(argv[i], "--"))
		fatal("improper end of names");

	ckargs("ABCERUP", argc, argv, i+1);

	if (get_cargval('A', &filename))
		nann = read_annotations(filename, annotation);

	/* open the file where bookmarks are to be written */
	if (get_cargval('B', &filename))
		bp = ckopen(filename, "w");
	else
		bp = stdout;

	if (get_cargval('C', &filename))
		ncon = read_conserved(filename, conserved);

	if (get_cargval('E', &filename))
		nexon = read_exons(filename, exon, gene, &ngene);

	if (get_cargval('R', &filename))
		nrpt = read_repeats(filename, repeat);

	if (get_cargval('U', &filename))
		nund = read_underlay(filename, underlay);

	/* output begins with the contents of the procset file */
	if (get_cargval('P', &procset))
	    cat(procset);
	else
	    cat(PREFIX "/lib/maps-procset.ps");

	show_align(mseq, nseq, annotation, nann, gene, ngene, exon, nexon,
	  repeat, nrpt, underlay, nund, conserved, ncon, bp);

	if (bp != stdout)
		fclose(bp);
	return 0;
}
